/*
**
*/
#include "InFBuf.h"

#define PROTECTED_IF(s, c, x) \
	register bool f; \
	acquire_##s; \
	(f = (c)) && (x);\
	release_##s; \
	if (f) 


//
// reader interface
//

const char PRIV_NAME[] = "(non-published area)";

// constructor

CInFBuf::CInFBuf(long size)
{
	long rsize = pageceil(size);
	long nrblocks = rsize/BLOCK_SIZE;

	long infosize = sizeof(FBufInfo);
	infosize += 2*nrblocks*sizeof(Block);
	infosize = pageceil(infosize);
	
#ifdef DEBUG
	printf("infosize:\t%d\n", infosize);
#endif
	
	infoarea = create_area(PRIV_NAME,
			&infoptr, B_ANY_ADDRESS,
			infosize,
			B_NO_LOCK, B_READ_AREA | B_WRITE_AREA);
	
	if (infoarea < B_NO_ERROR) {
		// some nicer way to handle this?
		fprintf(stderr, "Create FBuf failed\n");
		exit(1);
	}

	dataarea = B_ERROR;

	info = (FBufInfo *)infoptr;
	nextsegment = (Block *)(infoptr + sizeof(FBufInfo));
	count = (Block *)(nextsegment + nrblocks);
	
	info->first_free = 0;
	info->first_write = -1;
	info->last_write = -1;
#ifdef ALLOCLIST		
	info->first_alloc = -1;
#endif
	info->nrblocks = nrblocks;
	info->org_dataarea = -1;

#ifdef DEBUG
	fprintf(stderr, "info:\t%X\norg:\t%X\n", info, &info->org_dataarea);
	fprintf(stderr, "nextsegment:\t%X\n", nextsegment);
#endif	

	nextsegment[0] = -1;
	count[0] = nrblocks;

	// create semaphores/benaphores

	create_benaphore(&info->benaphore, "FBuf info");
	info->read_semaphore = create_sem(0, "FBuf read");
	info->alloc_semaphore = create_sem(0, "FBuf alloc");
	last_read = -1;
}

/*
CInFBuf::~CInFBuf()
{
}
*/

long	CInFBuf::Publish(const char *name)
{
	// first create the real shared buffer area
	dataarea = create_area("(FBuf data)",
			&dataptr, B_ANY_ADDRESS,
			info->nrblocks*BLOCK_SIZE,
			B_NO_LOCK, B_READ_AREA | B_WRITE_AREA);
	info->org_dataarea = dataarea; // for cloning

#ifdef DEBUG
	fprintf(stderr, "org_dataarea = %d\n", dataarea);
#endif
	
	if (dataarea < B_NO_ERROR) {
		return dataarea;
	}

	// make everything public by changing name of the info area
	area_id pubarea = clone_area(name,
			&infoptr, B_ANY_ADDRESS,
			B_READ_AREA|B_WRITE_AREA, infoarea);

#ifdef DEBUG
	fprintf(stderr, "infoarea = %d\n", infoarea);
	fprintf(stderr, "pubarea = %d\n", pubarea);
#endif 

	if (pubarea < B_NO_ERROR) {
		return pubarea;
	}
			
	// base addr has now new value
	// delete original area reference
	delete_area(infoarea);

	// update all local ptr references
		
	infoarea = pubarea;
	info = (FBufInfo *)infoptr;
	nextsegment = (Block *)(infoptr + sizeof(FBufInfo));
	count = (Block *)(nextsegment + info->nrblocks);

	return B_NO_ERROR;
}

long CInFBuf::Receive(void **buf)
{
	// read as much as possible from the write list

	Block block;

	// check if there is anything to read
	for (;;) {
		acquire_benaphore(info->benaphore);
		block = info->first_write;
#ifdef NOTYET
		// check if fbuf has grown
		if (info->nrblocks > nrblocks) {
			// grow our side accordingly
			Grow((info->nrblocks - nrblocks)*BLOCK_SIZE);
		}
#endif
		release_benaphore(info->benaphore);
		
		if (block >= 0)
			break;
			
		// nothing to read, wait for writer
		if (acquire_sem(info->read_semaphore) != B_NO_ERROR) {
			// this doesn't happen
			return B_ERROR;
		}
	}		

	info->first_write = nextsegment[block]; // only place where first_write may be assigned to -1

	*buf = dataptr + block*BLOCK_SIZE;
	return count[block]*BLOCK_SIZE;
}


void CInFBuf::Dispose(void *buf)
{
	// dispose back the buffer to the free-list

	Block block = ceilblock((char *)buf - dataptr);
		
	acquire_benaphore(info->benaphore);

#ifdef DEBUG
	fprintf(stderr, "Dispose()\tfirst_free: %d\tblock: %d\n", info->first_free, block);
#endif

	nextsegment[block] = info->first_free;
	info->first_free = block;

#ifdef DEBUG
	PrintToStream();
#endif

	if (info->requested_space > 0 && info->requested_space <= count[block]) {
		// signal the writer that there might be enough free blocks now
		info->requested_space = 0;
		release_benaphore(info->benaphore);
		release_sem(info->alloc_semaphore);
	}
	else {
		release_benaphore(info->benaphore);
	}		
}
