/* OutFBuf.cpp
** writer side
*/
#include "OutFBuf.h"


#define PROTECTED_IF(s, c, x) \
	register bool f; \
	acquire_##s; \
	(f = (c)) && (x);\
	release_##s; \
	if (f) 

COutFBuf::COutFBuf()
{
	info = 0;
	infoarea = -1;
	dataarea = -1;
}

/*
COutFBuf::~COutFBuf()
{
}
*/

long COutFBuf::Subscribe(const char *name)
{
	infoarea = find_area(name);
	if (infoarea < B_NO_ERROR) {
#ifdef DEBUG
		fprintf(stderr, "find_area('%s') failed\n", name);
#endif
		return infoarea;
	}
	
#ifdef DEBUG
	fprintf(stderr, "found info area = %d\n", infoarea);
#endif
	
	infoarea = clone_area("(cloned fbuf info)", 
			&infoptr,
			B_ANY_ADDRESS,
			B_READ_AREA | B_WRITE_AREA,
			infoarea);
#ifdef DEBUG
	fprintf(stderr, "cloned info area = %d\n", infoarea);
#endif

	if (infoarea < B_NO_ERROR) {
		return infoarea;
	}
	
	info = (FBufInfo *)infoptr;
	nextsegment = (Block *)(infoptr + sizeof(FBufInfo));
	count = (Block *)(nextsegment + info->nrblocks);

#ifdef DEBUG
	fprintf(stderr, "info:\t%X\norg:\t%X\n", info, &info->org_dataarea);
	fprintf(stderr, "nextsegment:\t%X\n", nextsegment);
#endif
	
	dataarea = clone_area("cloned data", 
			&dataptr,
			B_ANY_ADDRESS,
			B_READ_AREA | B_WRITE_AREA,
			info->org_dataarea);
	if (dataarea < B_NO_ERROR) {
#ifdef DEBUG
		fprintf(stderr, "clone data area %d failed\n", info->org_dataarea);
#endif
//		delete_area(info->org_dataarea); // TEMPORARY
		delete_area(infoarea);
		return dataarea;
	}

	return B_NO_ERROR;
}
	
#ifdef NOTYET
long COutFBuf::Resize(long size)
{
	size = pageceil(size);

	if (resize_area(dataarea, size) == B_ERROR) {
		// some other area is probably in the way
		fprintf(stderr, "resize_area() failed, trying to fix\n");
		// what to do now? should avoid moving things
		// create a new data area and link?
		
}
#endif


// must have size = n*BLOCK_SIZE for now!
void	*COutFBuf::Alloc(size_t size, bool wait, long flags, double timeout)
{
	long nrblocks = ceilblock(size);
	
	// find nrblocks continous free blocks
retry:
	acquire_benaphore(info->benaphore);
	Block i = info->first_free;
	release_benaphore(info->benaphore);
	Block from = -1;

	while ((i >= 0) && (count[i] < nrblocks)) {
		// not enough blocks at this segment, search further
		from = i;
		i = nextsegment[from];
	}
	if (i < 0) {
		if (wait) {
#ifdef DEBUG
			fprintf(stderr, "Alloc(): waiting for dispose()\n");
#endif
			acquire_benaphore(info->benaphore);
			info->requested_space = nrblocks;
			release_benaphore(info->benaphore);
			if (acquire_sem_etc(info->alloc_semaphore, 1, flags, timeout) == B_NO_ERROR) {
				goto retry;
			}
#ifdef DEBUG
			fprintf(stderr, "Alloc(): timeout or error\n");
#endif
		}						
		return 0;
	}
//	else {
		Block j;
		// allocate them
#ifdef DEBUG
		fprintf(stderr, "count[%d] == %d\tnrblocks == %d\n", i, count[i], nrblocks);
#endif
		if (count[i] == nrblocks) {
			// take whole segment
			j = nextsegment[i]; // j points to next segment
		}
		else {
			j = i + nrblocks; // j points to after the allocation
			// segment j replaces segment i in the list
			nextsegment[j] = nextsegment[i];
			// set correct count
			count[j] = count[i] - nrblocks; // what we took
			count[i] = nrblocks;
		}
		
		// special case first item
		{	PROTECTED_IF (benaphore(info->benaphore),
					i == info->first_free,
					info->first_free = j) {
			}
			else {
				nextsegment[from] = j;
			}
		}	


#ifdef ALLOCLIST
		// put the allocated item into the allocated list
		nextsegment[i] = info->first_alloc;
		info->first_alloc = i;
#endif		
		// return the allocated buffer
		return dataptr + i*BLOCK_SIZE;
}


long	COutFBuf::Send(void *buf)
{
	// put the buffer into the write list (queue)

	// which start block
	long offset = (char *)buf - dataptr;
	Block block = offset >> BLOCK_SIZE_SHIFT;
	long blockoffs = offset & (BLOCK_SIZE-1);
//	Block nrblocks = ceilblocks(size);
	Block nrblocks = count[block];
	
	// for now assume blockoffs == 0

#ifdef ALLOCLIST	
	// remove it from the allocate list
	if (info->first_alloc == block) {
		info->first_alloc = nextsegment[block];
	}
	else {
		// have to search the previous
		Block from, i = info->first_alloc;
		do {
			from = i;
			i = nextsegment[from];
		} while ((i >= 0) && (i != block));
		if (i >= 0) {
			nextsegment[from] = nextsegment[block]; // link over
		}
	}
#endif				
	
	{	PROTECTED_IF (benaphore(info->benaphore),
			(info->first_write >= 0), 0) {
			nextsegment[info->last_write] = block;
		}
	}
	nextsegment[block] = -1;
	count[block] = nrblocks;
	info->last_write = block;
	
	{	PROTECTED_IF (benaphore(info->benaphore),
				info->first_write < 0,
				info->first_write = block) {
			release_sem(info->read_semaphore);
		}
	}

	return nrblocks*BLOCK_SIZE; // - blockoffs;  // NOT CORRECT
}