/* FBuf.h
**
** Common base class for writer and reader (abstract)
**
** Fast non-copy IPC using shared memory.
** Inspired by
**	"Efficient Zero-Copy I/O Framework for UNIX"
**	Moti N. Thadani
**	Yousef A. Khalidi
**	SMLI TR-95-39, May 1995
**
** Revision list:
**	version 0.01a,		961019	/Johannes Elg, elg@df.lth.se
**		proof-of-concept
**		unidirectional
**	
**
*/
#pragma once

#include <OS.h>
#include <benaphore.h>

const BLOCK_SIZE_SHIFT = 7; 
const BLOCK_SIZE = 1 << BLOCK_SIZE_SHIFT;

class CFBuf {
protected:

	typedef char Block;

	// shared control info
	struct FBufInfo {
		benaphore_t benaphore; 
		sem_id	read_semaphore;
		sem_id	alloc_semaphore;
		area_id	org_dataarea;	// the writers area_id for the data
		Block	first_free, first_write;
		Block	last_write;
		Block	nrblocks;
		Block	requested_space;
#ifdef ALLOCLIST
		Block	first_alloc;
		Block	dummy[2]; // size of struct must be n*32 bits!
#else
		Block	dummy[3];
#endif
	};

// OBS: reader and writer has their own versions of these:

	char		*infoptr;
	char		*dataptr;
	area_id		infoarea;
	area_id		dataarea;

// constant pointers into the shared area:
	FBufInfo	*info;
	Block		*nextsegment;
	Block		*count;	

	CFBuf()
	{
		infoarea = dataarea = -1;
		info = 0;
		nextsegment = count = 0;
	}
	~CFBuf();
	
public:


	inline long ceilblock(long n)
	{
		return (n + BLOCK_SIZE-1) >> BLOCK_SIZE_SHIFT;
	}
	inline long pageceil(long n)
	{
		return (n + B_PAGE_SIZE-1) & ~(B_PAGE_SIZE-1);
	}

	void	PrintList(Block first);
	void	PrintToStream();
};